import os
import requests
from pkg_resources import resource_filename

def ensure_file_exists(file_dir, file_name, download_url):
    """
    Ensure the specified file exists. If it doesn't, download it from the provided URL.

    :param file_dir: The directory where the file should be located.
    :param file_name: The name of the file to check or download.
    :param download_url: The URL to download the file from.
    """
    # Combine directory and file name to create the full file path
    file_path = os.path.join(file_dir, file_name)

    # Check if the file already exists
    if os.path.exists(file_path):
        print(f"File already exists: {file_path}")
    else:
        print(f"File does not exist. Downloading from: {download_url}")
        
        # Ensure the directory exists
        os.makedirs(file_dir, exist_ok=True)
        
        # Download the file
        try:
            response = requests.get(download_url, stream=True)
            response.raise_for_status()  # Raise an exception for HTTP errors
            
            with open(file_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
            print(f"File successfully downloaded and saved to: {file_path}")
        except requests.exceptions.RequestException as e:
            print(f"File download failed: {e}")
            raise




import os
from pathlib import Path
import warnings
import pandas as pd

def check_paths(output_folder,output_prefix=None):
    # Create relative path
    output_path = os.path.join(os.getcwd(), output_folder)

    # Make sure that the folder exists
    Path(output_path).mkdir(parents=True, exist_ok=True)

    if os.path.exists(os.path.join(output_path, f"{output_prefix}assigned_locations.csv")):
        print("\033[91mWARNING\033[0m: Running this will overwrite previous results, choose a new"
              " 'output_folder' or 'output_prefix'")

    return output_path

def read_file(file_path, file_label):
    """
    Read data with given path.

    Parameters
    ----------
        file_path
            file path.
        file_label
            the file label to raise exception.

    Returns
    -------
        the read file.

    """
    try:
        file_delim = "," if file_path.endswith(".csv") else "\t"
        with warnings.catch_warnings():
            file_data = pd.read_csv(file_path, sep=file_delim).dropna()
    except Exception as e:
        raise IOError (f"Make sure you provided the correct path to {file_label} files. "
                    "The following input file formats are supported: .csv with comma ',' as "
                    "delimiter, .txt or .tsv with tab '\\t' as delimiter.")
    return file_data







import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from ._read_data import check_paths

def compute_cluster_averages(adata, annotation_key, common_cell,use_raw=False,project='',save=False,out_dir='./'):
    """
    Compute average expression of each gene in each cluster.

    Parameters
    ----------
    adata
        AnnData object of reference single-cell dataset.
    annotation_key
        Name of adata.obs column containing cluster labels.
    common_cell
        List to store the cell type order.

    Returns
    -------
    pd.DataFrame of cell type average expression of each gene

    """


    if not use_raw:
        x = adata.X
        var_names = adata.var_names
    else:
        if not adata.raw:
            raise ValueError("AnnData object has no raw data, change `use_raw=True, layer=None` or fix your object")
        x = adata.raw.X
        var_names = adata.raw.var_names

    averages_mat = np.zeros((1, x.shape[1]))

    for c in common_cell:
        sparse_subset = csr_matrix(x[np.isin(adata.obs[annotation_key], c), :])
        aver = sparse_subset.mean(0)
        averages_mat = np.concatenate((averages_mat, aver))
    averages_mat = averages_mat[1:, :].T
    df = pd.DataFrame(data=averages_mat, index=var_names, columns=common_cell)
    
    if save:
    # check out path
        reference_out_dir = check_paths(out_dir+'/reference_bulk_data')
        print('Saving average expression data')
        df.to_csv(reference_out_dir+f"/{project}_average_celltype_exp.txt",sep='\t')

    return df


def compute_bulk_with_average_exp(pseudo_bulk, average_cell_exp,save=False,out_dir='./',project=''):
    """
    Compute average expression of each gene in each cluster

    Parameters
    ----------
    pseudo_bulk
        AnnData object of reference single-cell dataset
    annotation_key
        Name of adata.obs column containing cluster labels
    common_cell
        List to store the cell type order.

    Returns
    -------
    pd.DataFrame of cell type average expression of each gene

    """
    pseudo_prop = pseudo_bulk.obs
    cell_sort = average_cell_exp.columns
    # resort prop data
    pseudo_prop = pseudo_prop[cell_sort]
    # truth bulk
    sample = np.zeros((pseudo_prop.shape[0],average_cell_exp.shape[0]))
    for i in range(pseudo_prop.shape[0]):
        cell_exp = pseudo_prop.iloc[i,:] * average_cell_exp
        sample[i] = cell_exp.sum(axis=1)
    # format dataframe
    sample = pd.DataFrame(sample,index=pseudo_bulk.obs_names,columns=pseudo_bulk.var_names)
    if save:
        reference_out_dir = check_paths(out_dir+'/reference_bulk_data')
        sample.to_csv(reference_out_dir+f"/{project}_cell_type_ave_bulk.txt",sep='\t')

    return sample

def data_dict_integration(common_list,data_df,data_dict,common_cell,top_num=100):
    """
    Integration the dataframe and dict.

    Parameters
    ----------
    data
        Dataframe and the columns is the key.
    dict
        Dictionary with key and values.
    top_number
        The top highest value will be keeped.

    Returns
    -------
    pd.DataFrame of cell type average expression of each gene.

    """
    key_list = data_dict.keys()
    overlapping_gene=[]
    for i in key_list:
        # find marker gene name in each cell type
        tmp_index = data_df[i].sort_values().iloc[:top_num].keys()
        if data_dict[i]!=[]:
            common_index = np.union1d(np.array(tmp_index), np.array(data_dict[i]))
        else:
            common_index = np.array(tmp_index)
        common_index = np.intersect1d(np.array(common_list),common_index)
        if len(common_index)>0:
            data_dict[i] = list(common_index)
            overlapping_gene += data_dict[i]
        else:
            del data_dict[i]
            common_cell.remove(i)
    overlapping_gene = np.unique(np.array(overlapping_gene))
    return data_dict, overlapping_gene, common_cell

def marker_integration(common_list,data_df,marker_list):
    """
    Integration the dataframe and dict.

    Parameters
    ----------
    data
        Dataframe and the columns is the key.
    dict
        Dictionary with key and values.
    top_number
        The top highest value will be keeped.

    Returns
    -------
    pd.DataFrame of cell type average expression of each gene.

    """

    overlapping_gene=[]
    data_df["mean"] = data_df.mean(axis=1)
    for i in data_df.columns:
        # find marker gene name in each cell type
        tmp_index = data_df.loc[data_df[i]>data_df["mean"]].index.tolist()
        overlapping_gene += tmp_index
    if len(marker_list)>0:   
        common_gene = np.union1d(overlapping_gene, marker_list)
    else:
        common_gene=np.array(overlapping_gene)
    common_gene = np.intersect1d(np.array(common_list),common_gene)

    return common_gene
    '''
    key_list = data_dict.keys()
    overlapping_gene=[]
    data_df["mean"] = data_df.mean(axis=1)
    for i in key_list:
        # find marker gene name in each cell type
        tmp_index = data_df.loc[data_df[i]>data_df["mean"]].index.tolist()
        if len(tmp_index)!=0:
            if len(data_dict[i])!=0:
                common_index = np.union1d(np.array(tmp_index), np.array(data_dict[i]))
            else:
                common_index = np.array(tmp_index)
        else:
            common_index = np.array(data_dict[i])
        common_index = np.intersect1d(np.array(common_list),common_index)
        if len(common_index)>0:
            data_dict[i] = list(common_index)
            overlapping_gene += data_dict[i]
        else:
            del data_dict[i]
            common_cell.remove(i)
    overlapping_gene = np.unique(np.array(overlapping_gene))
    return data_dict, overlapping_gene, common_cell
    '''

def filter_samples(pseudo_bulk, bulk_adata,data_num,num=20,cut_off=0.9,loc=None):
    from sklearn.metrics.pairwise import cosine_similarity
    sample_name = pseudo_bulk.obs_names
    if loc:
        bulk_matrix = bulk_adata.layers[loc]
        pseudo_matrix = pseudo_bulk.layers[loc]
    else:
        bulk_matrix = bulk_adata.X
        pseudo_matrix = pseudo_bulk.X
    #similarity_matrix = np.matmul(bulk_matrix,pseudo_matrix.T)
    similarity_matrix = cosine_similarity(bulk_matrix,pseudo_matrix)
    top_k_indices = np.argsort(-similarity_matrix, axis=1)[:, :num]
    selected_indices = np.unique(top_k_indices.flatten())
    if 5*similarity_matrix.shape[0]<2000:
        target_num = 2000
    else:
        target_num=5*similarity_matrix.shape[0]
    while len(selected_indices)<target_num:
          num=int(num*1.5)
          top_k_indices = np.argsort(-similarity_matrix, axis=1)[:, :num]
          selected_indices = np.unique(top_k_indices.flatten())
    return sample_name[selected_indices]

def compute_average_cosin(pseudo_bulk_tensor,bulk_adata,loc=None):
    from sklearn.metrics.pairwise import cosine_similarity
    valid_list=[]
    for i in pseudo_bulk_tensor:
        valid_list.append(i[0].numpy())
    pseudo_matrix = np.array(valid_list)
    if loc:
        bulk_matrix = bulk_adata.layers[loc]
    else:
        bulk_matrix = bulk_adata.X
    sample_cosine = cosine_similarity(pseudo_matrix,bulk_matrix)
    cosin_max = np.mean(sample_cosine, axis=1)
    return np.average(cosin_max)


def filter_gene(expression,reference,out_dir,cell_type,save=True):
    data_list=[expression,reference]
    filtered_list=[]
    for df in data_list:
        zero_count_per_column = (df == 0).sum(axis=0)
        columns_to_keep = zero_count_per_column[zero_count_per_column <= (len(df) / 2)].index
        df_filtered = df[columns_to_keep]
        filtered_list.append(df_filtered)
    common_columns = filtered_list[0].columns.intersection(filtered_list[1].columns)
    if save==True:
        reference_out_dir = check_paths(out_dir+'/test_data')
        reference[common_columns].to_csv(reference_out_dir+f"/{cell_type}.txt",sep='\t')
        

    return expression[common_columns],reference[common_columns]



def normalize_data(data):
    data = np.nan_to_num(data).astype(float)
    data *= 10**6 / np.sum(data, axis=0, dtype=float)
    np.log2(data + 1, out=data)
    np.nan_to_num(data, copy=False)
    return data



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
#import scanpy as sc
from .. import utils
import matplotlib.patches as mpatches


class Const:
    """
    Some COLOR SET used in the class.
    
    """
    COLOR_BLUE = ['#191970','#ADD8E6','#22AAA1']
    DIFF_COLOR_GRAY = ['#474350','#FAFAC6']
    FIG_FORMAT = 'svg'



def plot_mse_box(ground_truth,df_list,name_list,title,num_column=6,skipping_mse_calculation=False,
                color_set=Const.COLOR_BLUE,fig_format=None,out_dir='/out',filter_list=None,save=True):
    
    sns.set(font_scale=0.8,style="white")
    if skipping_mse_calculation:
        mse_result_list= df_list
    else:
        mse_result_list=[]
        for method_index in range(len(name_list)):
            df = df_list[method_index]
            mse_result = utils.eval_each_sample_mse(ground_truth,df,name_list[method_index])
            mse_result_list.append(mse_result)
    mse_df = pd.concat(mse_result_list)
    cell_name = ground_truth.columns.tolist()
    n_rows = int(len(cell_name)/num_column)+1
    fig, axes = plt.subplots(n_rows, num_column, sharex=True,sharey=False, figsize=(8,20))
    fig.suptitle(title,y=0.9)
    color_panel = sns.set_palette(color_set)
    max_y=[]
    if filter_list is None:
        filter_list = [2]*n_rows
    for i in range(n_rows):
        for j in range(num_column):
            if i*num_column+j<len(cell_name):
                sub_data = mse_df.loc[mse_df['cell_type'] == cell_name[i*num_column+j],"mse"]
                q1 = np.percentile(sub_data, 25)            
                q3 = np.percentile(sub_data, 80)            
                iqr = q3 - q1
                sub_data = sub_data[sub_data<q3+filter_list[i]*iqr]
                max_y.append(max(sub_data))
    for i in range(n_rows):
        each_row_y = max(max_y[i*num_column:(i+1)*num_column-1])
        for j in range(num_column):
            if i*num_column+j<len(cell_name):
                sub_data = mse_df.loc[mse_df['cell_type'] == cell_name[i*num_column+j],:]
                sns.boxplot(data=sub_data, x='method', y='mse',palette=color_panel,showfliers=False,width=.5,ax=axes[i, j])
                # rename for special cells
                if cell_name[i*num_column+j]=="T CD8 recently activated effector memory":
                    axes[i, j].set_title("T CD8 recently activated EM cells")
                if cell_name[i*num_column+j]=="pDC":
                    axes[i, j].set_title("pDC cells")
                else:
                    axes[i, j].set_title(cell_name[i*num_column+j])
                axes[i, j].set_xlabel("")
                axes[i, j].set_ylabel("")
                # reset the ylim
                if i==8:
                    axes[i, j].set_ylim(0,0.075)
                else:
                    axes[i, j].set_ylim(0,each_row_y)
                if j==0:
                    axes[i, j].set_yticklabels(axes[i, j].get_yticklabels())
                else:
                    axes[i, j].set_yticklabels("")
            else:
                custom_lines = []
                axes[i, j].set_xlabel("")
                axes[i, j].set_ylabel("")
                for m_index in method_index:
                    tmp_legand = mpatches.Rectangle([0,0],color=color_set[m_index],height=20,width=20,label=name_list[m_index],lw=0.5,edgecolor="black")
                    custom_lines.append(tmp_legand)
                axes[i,j].legend(handles=custom_lines, loc='upper center',bbox_to_anchor=(0.5,0.7),edgecolor='w',ncols=1)
                axes[i,j].axis('off')
    plt.edgecolor='white'
    if save:
        plt.savefig(f'{out_dir}/MSE_plot.{fig_format}', bbox_inches='tight')



import pandas as pd
import anndata as ad
import numpy as np
import scanpy as sc
from scipy.sparse import issparse,csr_matrix
from .. import utils
import time


#human_sc min_counts_per_cell=50 min_genes_per_cell=100
#hnsc min_counts_per_cell=1 min_genes_per_cell=5
def _filter_adata(
    adata,
    min_counts_per_gene=None,
    min_counts_per_cell=10,
    min_cells_per_gene=1,
    min_genes_per_cell=100,
    remove_constant_genes=True,
    remove_zero_cells=True,
    save=False,
    out_dir='.',
    project='',
    ignore_duplicated_genes=True
    ):

    """
    Filter one or more :class:`~anndata.AnnData` to satisfy simple quality criteria. 

    Parameters
    ----------
    adata
        An :class:`~anndata.AnnData` containing the expression to filter. 
    min_counts_per_gene: int, optional
        The minimum count (per :class:`~anndata.AnnData`) genes must have to be kept.
    min_counts_per_cell: int, optional
        The minimum count (per :class:`~anndata.AnnData`) cells must have to be kept.
    min_cells_per_gene: int, optional
        The minimum number of cells (per :class:`~anndata.AnnData`) genes must have to be kept.
    min_genes_per_cell: int, optional
        The minimum number of genes (per :class:`~anndata.AnnData`) cells must have to be kept.
    remove_constant_genes: boolean, optional
        Whether to remove genes which do not show any variation between cells
    remove_zero_cells: boolean, optional
        Whether to remove cells without non-zero genes.
        
    Returns
    -------
    Returns the filtered `adata`.
    
    """

    #check gene
    if not ignore_duplicated_genes:
        start_t = time.perf_counter()
        if not isinstance(adata.X, np.ndarray):
            adata.X = adata.X.toarray()
        print("Drop duplicated genes")
        #generate new data, average duplicated genes
        df_data = pd.DataFrame(adata.X,index=adata.obs_names,columns=adata.var_names).transpose()
        df_data['GeneSymbol'] = df_data.index
        df_data['GeneSymbol'] = df_data['GeneSymbol'].apply(lambda x: x.split('.')[0])
        df_data = df_data.groupby('GeneSymbol').mean()
        df_data = df_data.reset_index()
        df_data = pd.DataFrame(df_data.values[:,1:],index=df_data['GeneSymbol'].values,columns=df_data.columns[1:])
        meta = adata.obs
        cell_id = meta.index.tolist()
        common_id = list(set(cell_id) & set(df_data.columns))
        df_data = df_data.loc[:,common_id]
        meta = meta.loc[common_id,:]
        meta = meta.loc[df_data.columns,:]
        print(f"Found {df_data.shape[0]} not duplicated genes")
        X = np.array(df_data.values.T,dtype=float)
        X = csr_matrix(X)
        adata = sc.AnnData(X=X,obs=meta)
        adata.var_names = df_data.index
        print(f'Time to check the gene name and average the expression over duplicated genes: {round(time.perf_counter() - start_t, 2)} seconds')
    
    print("Start filtering")
    start_t = time.perf_counter()
    valid_genes = adata.var.index
    if (min_counts_per_gene is None and min_cells_per_gene is None and not remove_constant_genes) or np.prod(adata.shape) == 0:
        valid_genes = valid_genes.intersection(adata.var.index)
    else:
        if min_counts_per_gene is not None:
            valid_genes = valid_genes.intersection(adata.var.index[sc.pp.filter_genes(adata, min_counts=min_counts_per_gene, inplace=False)[0]])
        if min_cells_per_gene is not None:
            valid_genes = valid_genes.intersection(adata.var.index[sc.pp.filter_genes(adata, min_cells=min_cells_per_gene, inplace=False)[0]]) 
        if remove_constant_genes:
            if len(adata.obs) == 1:
                print('WARNING: Encountered a dataset with only a single observation while `remove_constant_genes` is enabled! As this would remove all genes, `remove_constant_genes` is disabled for this dataset.')
            else:
                not_constant = adata.X.max(axis=0)!=adata.X.min(axis=0)
                if issparse(not_constant):
                    not_constant = not_constant.todense().A
                not_constant = not_constant.flatten()
                valid_genes = valid_genes.intersection(adata.var.index[not_constant])
    if len(adata.var.index) != len(valid_genes): # filter happened
                adata = adata[:,valid_genes]
    elif (adata.var.index != valid_genes).any(): # reordering happened: no side effects on cell filtering
                adata = adata[:,valid_genes]
    ## check sample
    if np.prod(adata.shape) == 0:
        raise ValueError('Encountered a dataset with no high quality genes')

    cell_mask = np.full(len(adata.obs.index),True)
    if remove_zero_cells:
        cell_mask &= utils.get_sum(abs(adata.X),axis=1) != 0
    if min_counts_per_cell is not None:
        cell_mask &= sc.pp.filter_cells(adata, min_counts=min_counts_per_cell, inplace=False)[0]
    if min_genes_per_cell is not None:
        cell_mask &= sc.pp.filter_cells(adata, min_genes=min_genes_per_cell, inplace=False)[0]
    
    if not cell_mask.all():
        adata = adata[cell_mask]
    if np.prod(adata.shape) == 0:
        raise ValueError('Encountered a dataset with no valid samples or cells')
    
    print(f'The number of valid samples is {adata.shape[0]}, the number of valid genes is {adata.shape[1]}')

    print(f'Time to find subset samples which meet quality criterion: {round(time.perf_counter() - start_t, 2)} seconds')

    if save:
        out_dir = utils.check_paths(f'{out_dir}/filtered')
        adata.write_h5ad(f"{out_dir}/filtered_{project}.h5ad") 
    return adata


def _filter_dataframe(
    dataframe,
    min_counts_per_gene=None,
    min_counts_per_sample=100,
    min_cells_per_gene=None,
    min_genes_per_sample=100,
    remove_constant_genes=True,
    remove_zero_sample=True,
    project='',
    out_dir='.',
    **kwargs
    ):

    """
    Filter one or more :class:~pandas.dataframe to satisfy simple quality criteria. 

    Parameters
    ----------
    dataframe
        An :class:`~pandas.dataframe` containing the expression to filter. 
    min_counts_per_gene: int, optional
        The minimum count (per :class:`~pandas.dataframe`): genes must have to be kept.
    min_counts_per_sample: int, optional
        The minimum count (per :class:`~pandas.dataframe`): samples must have to be kept.
    min_samples_per_gene: int, optional
        The minimum number of samples (per :class:`~pandas.dataframe`): genes must have to be kept.
    min_genes_per_sample: int, optional
        The minimum number of genes (per :class:`~pandas.dataframe`) samplse must have to be kept.
    remove_constant_genes: boolean, optional
        Whether to remove genes which do not show any variation between samples.
    remove_zero_cells: boolean, optional: boolean, optional
        Whether to remove sample without non-zero genes.

    Returns
    -------
    Returns the filtered `adata`.
    
    """
    ## convert to `~anndata.AnnData`
    adata = sc.AnnData(dataframe)
    adata = _filter_adata(adata,
                        min_counts_per_gene=min_counts_per_gene,
                        min_counts_per_cell=min_counts_per_sample,
                        min_cells_per_gene=min_cells_per_gene,
                        min_genes_per_cell=min_genes_per_sample,
                        remove_constant_genes=remove_constant_genes,
                        remove_zero_cells=remove_zero_sample,
                        project=project,
                        out_dir=out_dir,
                        **kwargs)
    
    return adata

def qc_bulk_sc(bulk_data,
            sc_adata,
            min_counts_per_sample=50,
            min_genes_per_sample=100,
            out_dir='.',
            dataset_name='',
            **kwargs):
    """
    Quality control in bulk dataframe and sc adata.

    Parameters
    ----------
    bulk_data: dataframe
        An :class:`~pandas.dataframe` containing the bulk expression data. 
        The first column should be gene symbol, following column should be sample name.
    sc_adata: anndata.AnnData
        An :class:`~anndata.AnnData` containing the expression to filter.
    min_counts_per_sample: int, optional
        The minimum count (per :class:`~pandas.dataframe`): samples must have to be kept.
    min_genes_per_sample: int, optional
        The minimum number of genes (per :class:`~pandas.dataframe`) samplse must have to be kept.
    **kwargs:
        Additional keyword arguments forwarded to
        :func:`~cytobulk.preprocessing._filter_adata`.
    Returns
    -------
    Returns the filtered bulk data (adata) and sc data (adata).

    """
    sc_adata.var_names_make_unique()
    
    project_sc = dataset_name+'_sc'
    project_bulk = dataset_name+'_bulk'
    # check data format
    if not isinstance(sc_adata,ad.AnnData):
        raise ValueError(f"`A` can only be a dataframe but is a {sc_adata.__class__}!")
    if isinstance(bulk_data,pd.DataFrame):
        print("------------------start filtering bulk data-------------------------")
        filtered_bulk = _filter_dataframe(bulk_data,
                                        min_counts_per_sample=min_counts_per_sample,
                                        min_genes_per_sample=min_genes_per_sample,
                                        project = project_bulk,
                                        out_dir=out_dir,
                                        **kwargs)
        print("------------------finish filtering bulk data-------------------------")
    else:
        bulk_data.var_names_make_unique()
        filtered_bulk = _filter_adata(bulk_data,
                                    min_counts_per_cell=min_counts_per_sample,
                                    min_genes_per_cell=min_genes_per_sample,
                                        project = project_bulk,
                                        out_dir=out_dir,
                                        **kwargs)

    print("------------------start filtering single cell data-------------------------")
    filtered_sc = _filter_adata(sc_adata,
                                project = project_sc,
                                out_dir=out_dir,
                                **kwargs)
    print("------------------finish filtering single cell data-------------------------")
    # get overlapping gene between sc adata and bulk adata
    common_gene = filtered_bulk.var.index.intersection(filtered_sc.var.index)
    # subset datasets
    filtered_sc = filtered_sc[:,common_gene]
    filtered_bulk = filtered_bulk[:,common_gene]
    print("finish getting overlapping genes")
    return filtered_sc, filtered_bulk, common_gene

def high_variable_gene(adata,flavor):
    sc.pp.highly_variable_genes(adata, min_mean=0, max_mean=np.inf, min_disp=0.25,flavor=flavor)
    print("length")
    print(len(adata.var.highly_variable))
    if len(adata.var.highly_variable)>4000:
        sc.pp.highly_variable_genes(adata, min_mean=0, max_mean=np.inf, min_disp=0.5,flavor=flavor)
    if len(adata.var.highly_variable)>4000:
        sc.pp.highly_variable_genes(adata, n_top_genes=4000)
    return adata[:, adata.var.highly_variable]

def qc_st_sc(st_adata,
            sc_adata,
            min_counts_per_sample=1,
            min_genes_per_sample=5,
            dataset_name='',
            out_dir='.',
            **kwargs):
    """
    Quality control in bulk dataframe and sc adata.

    Parameters
    ----------
    bulk_data: dataframe
        An :class:`~pandas.dataframe` containing the bulk expression data. 
        The first column should be gene symbol, following column should be sample name.
    sc_adata: anndata.AnnData
        An :class:`~anndata.AnnData` containing the expression to filter.
    min_counts_per_sample: int, optional
        The minimum count (per :class:`~pandas.dataframe`): samples must have to be kept.
    min_genes_per_sample: int, optional
        The minimum number of genes (per :class:`~pandas.dataframe`) samplse must have to be kept.
    **kwargs:
        Additional keyword arguments forwarded to
        :func:`~cytobulk.preprocessing._filter_adata`.
    Returns
    -------
    Returns the filtered bulk data (adata) and sc data (adata).

    """
    sc_adata.var_names_make_unique()
    st_adata.var_names_make_unique()
    project_sc = dataset_name+'_sc'
    project_st = dataset_name+'_st'
    # check data format
    if not isinstance(sc_adata,ad.AnnData):
        raise ValueError(f"`A` can only be a dataframe but is a {sc_adata.__class__}!")
    if not isinstance(st_adata,ad.AnnData):
        raise ValueError(f"`A` can only be a anndata.AnnData but is a {st_adata.__class__}!")
    print("------------------start filtering st data-------------------------")
    filtered_st = _filter_adata(st_adata,
                                    min_counts_per_cell=min_counts_per_sample,
                                    min_genes_per_cell=min_genes_per_sample,
                                    min_cells_per_gene=10,
                                    project = project_st,
                                    out_dir= out_dir,
                                    **kwargs)
    print("------------------finish filtering st data-------------------------")

    print("------------------start filtering single cell data-------------------------")
    filtered_sc = _filter_adata(sc_adata,
                                project = project_sc,
                                out_dir=out_dir,
                                **kwargs)
    print("------------------finish filtering single cell data-------------------------")
    # get overlapping gene between sc adata and bulk adata
    common_gene = filtered_st.var.index.intersection(filtered_sc.var.index)
    # subset datasets
    filtered_sc = filtered_sc[:,common_gene]
    filtered_st = filtered_st[:,common_gene]
    print(f"finish getting {len(common_gene)} overlapping genes")
    return filtered_sc, filtered_st, common_gene

def qc_sc(sc_adata,**kwargs):
    """
    Quality control in sc adata.

    Parameters
    ----------
    sc_adata: anndata.AnnData
        An :class:`~anndata.AnnData` containing the expression to filter.
    **kwargs:
        Additional keyword arguments forwarded to
        :func:`~cytobulk.preprocessing._filter_adata`.
    Returns
    -------
    Returns the filtered sc data (adata).

    """

    if not isinstance(sc_adata,ad.AnnData):
        raise ValueError(f"`A` can only be a ad.AnnData but is a {sc_adata.__class__}!")
    print("------------------start filtering single cell data-------------------------")
    filtered_adata = _filter_adata(sc_adata,**kwargs)
    print("------------------finish filtering single cell data-------------------------")
    return filtered_adata


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

class Const:
    PIX_X = 'X'
    PIX_Y = 'Y'

def rgb2grey(img: np.ndarray):
    return np.dot(img[...,:3], [0.2989, 0.5870, 0.1140])

def predict_cell_num(st_adata,
                     library_ids="library_ids",
                     project="test",
                     diameter=None,
                     save_png_result=False,
                     model_type='cyto3',
                     out_dir='.',
                     cellprob_threshold=0.4,
                     save=True):
    
    '''
    Predicts cell numbers from spatial transcriptomics data using Cellpose.

    Parameters
    ----------
    st_adata : AnnData
        Spatial transcriptomics AnnData object.

    library_ids : string
        Identifier for the library in the AnnData object.

    diameter : int, optional
        Estimated diameter of the cells to be detected.

    save_png_result : bool, optional
        If True, saves the segmentation results as PNG files.

    model_type : string, optional
        Type of Cellpose model to use. Default is 'cyto3'.

    out_path : string, optional
        Directory to save the output files. Default is './'.

    cellprob_threshold : float, optional
        Cell probability threshold for Cellpose. Default is 1.

    save : bool, optional
        If True, saves the results to an h5ad file. Default is True.

    Returns
    -------
    st_adata : AnnData
        Updated AnnData object with predicted cell numbers.

    cell_pos : pandas.DataFrame
        DataFrame containing cell positions.
    '''
    from cellpose import models, io, plot
    from tqdm import tqdm

    print('-----Initializing model...')
    model = models.Cellpose(model_type=model_type)
    ch = [0, 0] # NOTE: here we set all images to greyscale

    print('-----Reading files...')
    img = rgb2grey(st_adata.uns['spatial'][library_ids]['images']['hires'])


    coord = st_adata.obsm['spatial']*st_adata.uns['spatial'][library_ids]['scalefactors']['tissue_hires_scalef']
    spots = pd.DataFrame(coord, columns=["X", "Y"])
    crop_r = int(st_adata.uns['spatial'][library_ids]['scalefactors']['spot_diameter_fullres']*st_adata.uns['spatial'][library_ids]['scalefactors']['tissue_hires_scalef'])
    #crop_r = int(st_adata.uns['spatial'][library_ids]['scalefactors']['spot_diameter_fullres'])
    half_r = crop_r // 2 + 5
    print(half_r)
    print('-----Predicting cell number...')
    ret = pd.DataFrame(data={'X':[], 'Y':[], 'cell_num':[]})
    cell_pos = pd.DataFrame(data={'id':[], 'X':[], 'Y':[]})
    for _, row in tqdm(spots.iterrows()):
        x = int(row[Const.PIX_X]); y = int(row[Const.PIX_Y])
        x_max = min(x+half_r, img.shape[0]-1)
        x_min = max(x-half_r, 0)
        y_max = min(y+half_r, img.shape[1]-1)
        y_min = max(y-half_r, 0)

        tile = img[x_min:x_max, y_min:y_max]
        masks, flows, styles, diams = model.eval(tile, diameter=diameter, channels=ch,  cellprob_threshold=cellprob_threshold)
        cell_num = len(np.unique(masks))
        ret.loc[len(ret.index)] = [x, y, cell_num]
        for i in range(cell_num):
            xi = np.where(masks == i)[0].mean()
            yi = np.where(masks == i)[1].mean()
            cell_pos.loc[len(cell_pos.index)] = [f"spot{_}_cell{i}", xi, yi]
        
        if save_png_result:
            fig = plt.figure()
            plot.show_segmentation(fig, tile, masks, flows[0], channels=ch)
            plt.tight_layout()
            #plt.savefig(save_png_result.replace('.', f'_{x}x{y}.'))
            plt.savefig(f"{out_dir}/figures/{_}_segmentation_result.png")

    st_adata.obsm["cell_num"] = (ret["cell_num"]).to_numpy()
    st_adata.uns["seg_cell_pos"] = cell_pos
    ret["cell_num"].to_csv(f"{out_dir}/output/{project}_cell_num.csv")
    cell_pos.to_csv(f"{out_dir}/output/{project}_cell_pos.csv")
    st_adata.write_h5ad(f"{out_dir}/output/segmentation_adata.h5ad")
    return st_adata, cell_pos



import torch
import pandas as pd
import numpy as np
import torch.nn as nn
import multiprocessing as mp
import json
import warnings
import sys
from tqdm import tqdm
from os.path import exists
from typing import Callable, Generator
import torch.distributed as dist
from scipy.stats import pearsonr
from ... import utils
from ... import get
from ... import plots
from ... import preprocessing
import pickle 


class Const:
    """
    Some constants used in the class.
    """
    MODE_TRAINING = "training"
    MODE_PREDICTION = "prediction"
    SAMPLE_COL = "sample_name"
    GENE_SYMBOL_COL = "GeneSymbol"
    BATCH_SIZE = 64
    LEARNING_RATE = 0.005
    MAX_SPLIT = 10
    EPOCH_NUM_BULK = 12
    EPOCH_NUM_ST = 10
    SEED = 20230602
    CHEB_MODE = 0
    MAX_RETRY=1


def configure_device(use_gpu):
    if use_gpu:
        if torch.cuda.is_available(): return "cuda"
        if torch.backends.mps.is_available(): return "mps"
    return "cpu"


def naive_parallel(
        func: Callable, 
        args: Generator, 
        cpu_limit = mp.cpu_count()
):
    print(f"Try to allocate {cpu_limit}, ", end='')
    cpu_limit = min(cpu_limit, mp.cpu_count())
    print(f"{cpu_limit} cpu(s) are currently available.")

    with mp.Pool(processes=cpu_limit) as pool:
        ret = pool.starmap_async(func, args)
        print(ret.get())


class LinearModel(torch.nn.Module):
    def __init__(self, t):
        super(LinearModel, self).__init__()
        self.encoder = torch.nn.Sequential(
            torch.nn.Linear(t,64),
            nn.BatchNorm1d(64),
            torch.nn.Linear(64,1),
            torch.nn.Sigmoid()
        )

    def forward(self, t):
        f = self.encoder(t)
        return f
    

class GraphConv(nn.Module):
    def __init__(self, in_c, out_c, K, device, bias=True, normalize=True):
        super(GraphConv, self).__init__()
        self.device = device
        self.normalize = normalize

        self.weight = nn.Parameter(torch.Tensor(K + 1, 1, in_c, out_c))
        nn.init.orthogonal_(self.weight, gain=nn.init.calculate_gain('leaky_relu', 0.4))

        if bias:
            self.bias = nn.Parameter(torch.Tensor(1, 1, out_c))
            # nn.init.zeros_(self.bias)
            nn.init.orthogonal_(self.bias, gain=nn.init.calculate_gain('leaky_relu', 0.4))
        else:
            self.register_parameter("bias", None)

        self.K = K + 1

    def forward(self, inputs, graph, mode=Const.CHEB_MODE):
        L = GraphConv.get_laplacian(graph, self.normalize)
        L = L.cpu().numpy()
        lam, u = np.linalg.eig(L)
        q1 = np.percentile(lam.real,50)
        #lam = torch.from_numpy(lam)
        #lam = lam.to(self.device)
        #lam= torch.diag(lam)
        #torch.set_printoptions(precision=3,sci_mode=False)
        lam=lam.real
        q1 = np.percentile(lam,50)
        lam[lam<q1]=0
        lam = torch.from_numpy(lam)
        #print(lam)
        lam= torch.diag(lam).to(self.device)
        u = torch.from_numpy(u).to(self.device)
        #u = u.to(self.device)
        lam = 2*((lam - torch.min(lam).to(self.device)) / (torch.max(lam).to(self.device) - torch.min(lam).to(self.device))) - torch.eye(lam.size(0)).to(self.device)
        mul_L = self.cheb_polynomial(lam).unsqueeze(1)

        if mode == 0:
            result = torch.matmul(inputs, mul_L)
        elif mode == 1:
            result = torch.matmul(inputs, mul_L)
            result = torch.matmul(result, u.t())
        elif mode == 2:
            result = torch.matmul(u, mul_L)
            result = torch.matmul(result, u.t())
            result = torch.matmul(inputs, result)

        result = torch.matmul(result, self.weight)
        result = torch.sum(result, dim=0) + self.bias

        temp=[]
        for i in range(result.size()[0]):
            if torch.min(result[i]) == torch.max(result[i]):
                temp.append(result[i])
            else:
                temp.append(2*((result[i] - torch.min(result[i])) / (torch.max(result[i]) - torch.min(result[i]))) - 1)
        temp = torch.stack(temp)

        return temp

    def cheb_polynomial(self, laplacian):
        N = laplacian.size(0)
        multi_order_laplacian = torch.zeros([self.K, N, N], device=laplacian.device, dtype=torch.float)
        multi_order_laplacian[0] = torch.eye(N, device=laplacian.device, dtype=torch.float)

        if self.K == 1:
            return multi_order_laplacian
        else:
            multi_order_laplacian[1] = laplacian
            if self.K == 2:
                return multi_order_laplacian
            else:
                for k in range(2, self.K):
                    multi_order_laplacian[k] = 2 * torch.mm(laplacian, multi_order_laplacian[k-1]) - multi_order_laplacian[k-2]

        return multi_order_laplacian

    @staticmethod
    def get_laplacian(graph, normalize):
        if normalize:

            D = torch.diag(torch.sum(graph, dim=-1) ** (-1 / 2))
            L = torch.eye(graph.size(0), device=graph.device, dtype=graph.dtype) - torch.mm(torch.mm(D, graph), D)
        else:
            D = torch.diag(torch.sum(graph, dim=-1))
            L = D - graph
        return L


class GraphNet_bulk(nn.Module):
    def __init__(self, in_c, hid_c, out_c, K, device):
        super(GraphNet_bulk, self).__init__()
        self.conv1 = GraphConv(in_c=in_c, out_c=hid_c, K=K, device=device)
        #self.conv2 = GraphConv(in_c=hid_c, out_c=out_c, K=K, device=device)
        self.act = nn.ELU()

    def forward(self, graph, data):
        graph_data = graph
        flow_x = data

        B, N = flow_x.size(0), flow_x.size(1)

        flow_x = flow_x.view(B, 1, N)

        output_1 = self.act(self.conv1(flow_x, graph_data))
        #output_2 = self.act(self.conv2(output_1, graph_data))

        return output_1
    

class GraphNet_st(nn.Module):
    def __init__(self, in_c, hid_c, out_c, K, device):
        super(GraphNet_st, self).__init__()
        self.conv1 = GraphConv(in_c=in_c, out_c=out_c, K=K, device=device)
        self.act = nn.ELU()

    def forward(self, graph, data):
        graph_data = graph
        flow_x = data

        B, N = flow_x.size(0), flow_x.size(1)

        flow_x = flow_x.view(B, 1, N)

        output_1 = self.act(self.conv1(flow_x, graph_data))
        #output_2 = self.act(self.conv2(output_1, graph_data))

        return output_1

def change_lr(optim, new_lr):
    for g in optim.param_groups:
        g['lr'] = new_lr

class InferDataset(torch.utils.data.Dataset):
    def __init__(self, data):
        self.data = data
        
    def __getitem__(self, index):
        x = self.data[index]
        return x
    
    def __len__(self):
        return len(self.data)

def get_G(cell_name, sc_adata,annotation_key):
        def _get_mat_YW(sc_df):
            mat_Y = torch.from_numpy(sc_df.values)
            mat_W = mat_Y @ mat_Y.t()
            return mat_W

        sec_num = 1e-20
        #sub_adata = sc_adata[sc_adata.obs[annotation_key]==cell_name,sel_gene].copy()
        sub_adata = sc_adata[sc_adata.obs[annotation_key]==cell_name,].copy()
        sub_df = get.count_data(sub_adata)
        mat_W = _get_mat_YW(sub_df)
        num = len(mat_W)
        mat_G = mat_W + sec_num*torch.eye(num) + sec_num*torch.ones(num, num)
        return mat_G, num

def select_gene(expression: pd.DataFrame, sel_gene: list):
    warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)
    ret_exp = pd.DataFrame(expression.iloc[:, 0])
    for gene in sel_gene:
        ret_exp[gene] = expression[gene] if gene in expression.columns else 0
    return ret_exp.iloc[:, 1:]

def train_cell_loop_once(cell, 
                        expression,
                        fraction,
                        bulk_adata,
                        batch_size,
                        sc_adata,
                        out_dir,
                        device,
                        annotation_key,
                        project_name,
                        data_num,
                        loc,
                        is_st):

    if exists(f'{out_dir}/graph_{cell}.pt') and exists(f'{out_dir}/linear_{cell}.pt'):
        print(f"skipping training model for {cell}")
    else:
        if is_st:
            meet_req = Const.MAX_RETRY
        else:
            meet_req = 1
        print(f"training model for {cell}")
        #sel_gene = marker[cell]
        #presudo_bulk_full = presudo_bulk[:,sel_gene].copy()
        #bulk_adata_full = bulk_adata[:,sel_gene].copy()
        #presudo_bulk_full = presudo_bulk.copy()
        bulk_adata_full = bulk_adata.copy()
        #presudo_bulk_full, bulk_adata_full = preprocessing.remove_batch_effect(presudo_bulk_full, bulk_adata_full, out_dir=out_dir, project=project_name+"_"+cell,batch_effect=batch_effect)

        #if batch_effect:
            #loc="batch_effected"
        #else:
            # loc=None
        #cut_off=0.975
        

        #sample_list = utils.filter_samples(presudo_bulk_full, bulk_adata_full,data_num=data_num,cut_off=cut_off,loc=loc)

        #presudo_bulk_train = presudo_bulk_full[sample_list,:]
        #print(presudo_bulk_train.shape)
        #expression = get.count_data_t(presudo_bulk_train,counts_location=loc)
        #reference = get.count_data_t(bulk_adata_full,counts_location=loc)
        #fraction = get.meta(presudo_bulk_train,position_key="obs")
        #print("batch_effect",batch_effect)
        #if batch_effect:
            #plots.batch_effect(bulk_adata_full, presudo_bulk_train,out_dir=out_dir+"/plot",title=cell)
        #print(f"Start training the model for {cell} cell type...")
        #sel_gene = expression.columns
        mat_G, num = get_G(cell, sc_adata,annotation_key)
        mat_G = mat_G.to(device)
        #input_bulk = expression[sel_gene]
        #bulk_adata_full = bulk_adata_full[:,sel_gene]
        #select_gene(expression, sel_gene)
        train_data = expression.values
        train_label = fraction[cell].values
        full_dataset = torch.utils.data.TensorDataset(torch.from_numpy(train_data), torch.from_numpy(train_label))
        train_size = int(batch_size * (0.85 * len(full_dataset) // batch_size))
        valid_size = len(full_dataset) - train_size      
        best_cosin = 0
        random_split=Const.MAX_SPLIT
        while random_split:
            random_split -= 1
            if random_split == Const.MAX_SPLIT-1:
                train_dataset, valid_dataset = torch.utils.data.random_split(full_dataset, [train_size, valid_size])
                best_cosin = utils.compute_average_cosin(valid_dataset,bulk_adata_full,loc=loc)
            else:
                _dataset, _valid = torch.utils.data.random_split(full_dataset, [train_size, valid_size])
                _cosin = utils.compute_average_cosin(_valid,bulk_adata_full,loc=loc)
                if _cosin > best_cosin:
                    train_dataset = _dataset
                    valid_dataset = _valid
                    best_cosin = _cosin

        train_loader = torch.utils.data.DataLoader(dataset = train_dataset, batch_size = batch_size, shuffle = True)
        valid_loader = torch.utils.data.DataLoader(dataset = valid_dataset, batch_size = 1, shuffle = False)
        max_retry=5
        while meet_req>0 and max_retry>0:
            meet_req -= 1
            max_retry -= 1
            smooth_times=0

            if not is_st:
                model_graph = GraphNet_bulk(num,num,num,2, device=device).to(device)
            else:
                model_graph = GraphNet_st(num,num,num,2, device=device).to(device)
            model_graph_optim = torch.optim.Adam(model_graph.parameters(), lr=Const.LEARNING_RATE, weight_decay = 1e-8)
            model_graph_schelr = torch.optim.lr_scheduler.StepLR(model_graph_optim, 5, gamma=0.9)
            model_linear = LinearModel(num).to(device)
            model_linear_optim = torch.optim.Adam(model_linear.parameters(), lr=Const.LEARNING_RATE, weight_decay = 1e-8)
            model_linear_schelr = torch.optim.lr_scheduler.StepLR(model_linear_optim, 5, gamma=0.9)

            plot_info_dict = {"mse_loss": []}
            if max_retry == 4: 
                pre_pearson_r = -np.inf
                pre_loss = np.inf
                model_r = pre_pearson_r
                model_loss = pre_loss
                best_graph = None; best_linear = None
            graph_break = False
            linear_stop = False
            # for epo in tqdm(range(Const.EPOCH_NUM), leave=False):
            if is_st:
                epoch_num = Const.EPOCH_NUM_ST
            else:
                epoch_num = Const.EPOCH_NUM_BULK
            for epo in range(epoch_num):
                model_graph.train(); model_linear.train()
                for _, (data, target) in enumerate(train_loader):
                    target = torch.reshape(target,(batch_size, 1))
                    data = data.to(torch.float32).to(device)
                    target = target.to(torch.float32).to(device)
                    model_graph_optim.zero_grad()
                    model_linear_optim.zero_grad()
                    zlist=torch.reshape(model_graph(mat_G, data), (batch_size, -1))      
                    output_frac = model_linear(zlist)
                    loss_f = ((output_frac-target)**2).sum() / 1 / batch_size
                    plot_info_dict["mse_loss"].append(loss_f.data.cpu().detach().clone().numpy().item())
                    loss_f.backward()

                    if not graph_break: model_graph_optim.step()
                    if not linear_stop: model_linear_optim.step()

                model_graph.eval(); model_linear.eval()
                with torch.no_grad():
                    valid_cor_dict = {
                        "frac_pred": [],
                        "frac_truth": []
                    }
                    for _, (data, target) in enumerate(valid_loader):
                        target = torch.reshape(target,(1, 1))
                        data = data.to(torch.float32).to(device)
                        target = target.to(torch.float32).to(device)

                        zlist = torch.reshape(model_graph(mat_G, data), (1, -1))      
                        output_frac = model_linear(zlist)
                        loss_f = ((output_frac-target)**2).sum()
                        valid_cor_dict["frac_pred"].append(output_frac.cpu().detach().clone().numpy().item())
                        valid_cor_dict["frac_truth"].append(target.cpu().detach().clone().numpy().item())
                    
                    pearson_r, pearson_p = pearsonr(valid_cor_dict["frac_pred"], valid_cor_dict["frac_truth"])
                    if epo==0 and max_retry == 4:
                        pre_loss = loss_f
                        pre_pearson_r = pearson_r
                        model_r = pearson_r
                        model_loss = loss_f
                        best_graph = model_graph.state_dict()
                        best_linear = model_linear.state_dict()
                    else:
                            if (pearson_r >= model_r-0.005 and model_loss > loss_f) or (model_r<0) or (pearson_r >= model_r and model_loss > loss_f-0.001):
                                model_loss = loss_f
                                model_r = pearson_r
                                best_graph = model_graph.state_dict()
                                best_linear = model_linear.state_dict()
                            #print(f"pearson_r = {model_r}...epo={epo}", end=" ")
                    if (abs(pre_pearson_r-pearson_r)>0.01) and (smooth_times==1):
                        smooth_times=0
                    if (abs(pre_pearson_r-pearson_r)<=0.01):
                        smooth_times+=1
                    '''
                    if is_st:
                    '''
                    if not graph_break and epo>0:
                        if (pearson_r> 0.97 and epo<=9) or ((abs(pre_pearson_r-pearson_r)<0.01) and (smooth_times==2)):
                            graph_break = True
                            change_lr(model_linear_optim, 0.0001)
                            print("stop graph training,linear training--0.0001")
                        elif (pearson_r> 0.95 and epo>3 and epo<=9) or ((abs(pre_pearson_r-pearson_r)<0.01) and (smooth_times==2)):
                            graph_break = True
                            change_lr(model_linear_optim, 0.0001)
                            print("stop graph training,linear training--0.0001")
                        elif (pearson_r> 0.85 and epo>5 and epo<=9) or ((abs(pre_pearson_r-pearson_r)<0.01) and (smooth_times==2)):
                            graph_break = True
                            change_lr(model_linear_optim, 0.0001)
                            print("graph training--0.0005,linear training--0.0015")
                        elif (pearson_r> 0.75 and epo>7 and epo<=9) or ((abs(pre_pearson_r-pearson_r)<0.01) and (smooth_times==2)):
                            graph_break = True
                            change_lr(model_linear_optim, 0.0001)
                            print("graph training--0.0005,linear training--0.0015")
                        elif epo> 9:
                            graph_break = True
                            change_lr(model_linear_optim, 0.0001)
                            print("epo>9,graph training break,linear training--0.0005")
                    elif graph_break and epo>0:
                        if not linear_stop:
                            if loss_f<0.001 or abs(pre_loss-loss_f) < 0.0005:
                                linear_stop = True
                            elif abs(pre_loss-loss_f) <0.01 and abs(pre_loss-loss_f) > 0.001:
                                change_lr(model_linear_optim, 0.001)
                                print("change linear training--0.001")
                    '''
                    else:
                        if not graph_break and epo>0:
                            if pearson_r> 0.95 and epo<=5:
                                graph_break = True
                                change_lr(model_linear_optim, 0.0001)
                                print("stop graph training,linear training--0.0005")
                            elif pearson_r> 0.9 and epo>5 and epo<=7:
                                graph_break = True
                                change_lr(model_linear_optim, 0.0001)
                                print("graph training--0.0005,linear training--0.0015")
                            elif pearson_r> 0.85 and epo>7 and epo<=10:
                                change_lr(model_graph_optim, 0.0001)
                                change_lr(model_linear_optim, 0.0015)
                                print("graph training--0.0005,linear training--0.0015")
                            elif epo>= 13:
                                graph_break = True
                                change_lr(model_linear_optim, 0.0005)
                                print("epo>15,graph training break,linear training--0.0005")
                        elif graph_break and epo>1:
                            if not linear_stop:
                                    if loss_f < 0.01 and loss_f > 0.005:
                                        change_lr(model_linear_optim, 0.0005)
                                        print("change linear training--0.0005")
                                    elif loss_f < pre_loss and pre_loss-loss_f <= 0.001:
                                        linear_stop = True
                                        print("stop linear training")
                                    elif loss_f <= 0.002:
                                        change_lr(model_linear_optim, 0.0001)
                                        print("change linear training--0.0001")
                    '''
                    if linear_stop and graph_break:
                        break

                    print(f"epoch{epo}-pearsonR", pearson_r, pearson_p, loss_f)
                    pre_pearson_r = pearson_r
                    pre_loss = loss_f
                    model_graph_schelr.step()
                    model_linear_schelr.step()
            with open(f'{out_dir}/plot/train_plot_info_{cell}.json', 'w') as f: 
                json.dump(plot_info_dict, f)
            print(f"Saving {cell} model...", end=" ")
            print(f"pearson_r = {model_r}...loss={model_loss}...", end=" ")
            # TODO: path check
            torch.save(best_graph, f"{out_dir}/graph_{cell}.pt")
            torch.save(best_linear, f"{out_dir}/linear_{cell}.pt")
            print("Done.")
            if model_r <0.90 or model_loss>0.01:
                meet_req+=1
    
            
class GraphDeconv:
    def __init__(
            self,
            cell_num=200,
            mode=Const.MODE_PREDICTION,
            use_gpu=True
    ):
        """
            cell_num: int, the number of cell for each bulk sample.
            mode: string, prediction or training.
            use_gpu: bool, if `True`, the model will use CUDA or MPS, otherwise, it will only use CPU.
        """
        self.cell_num = cell_num
        self.mode = mode
        self.device = configure_device(use_gpu)

    def fit(
        self,
        expression,
        cell_list=None,
        sc_adata=None,
        annotation_key = None,
        model_folder=None,
        out_dir='./',
        file_dir = "./",
        project='',
        save=True,
        is_st=False):
        
        utils.check_paths(output_folder=out_dir)
        device=self.device
        tot_cell_list = cell_list
        final_ret = pd.DataFrame()
        for cell in tqdm(tot_cell_list, leave=False):
            mat_G, num = get_G(cell, sc_adata,annotation_key)
            mat_G = mat_G.to(device)
            #input_bulk = expression[sel_gene]
            input_bulk = pd.read_csv(f"{file_dir}/batch_effect/{project}_batch_effected.txt",sep='\t',index_col=0)[expression.obs_names.tolist()]

            input_bulk.clip(lower=0,inplace=True)
            #input_bulk = input_bulk.T
            #test_data = input_bulk[sel_gene].values
            test_data = input_bulk.T.values


            test_dataset = InferDataset(torch.FloatTensor(test_data))
            test_loader = torch.utils.data.DataLoader(dataset = test_dataset, batch_size = 1, shuffle = False)

            if not is_st:
                model_graph = GraphNet_bulk(num,num,num,2, device=device).to(device)
            else:
                model_graph = GraphNet_st(num,num,num,2, device=device).to(device)
            
            model_graph.load_state_dict(torch.load(f'{model_folder}/graph_{cell}.pt'))
            model_linear = LinearModel(num).to(self.device)
            model_linear.load_state_dict(torch.load(f'{model_folder}/linear_{cell}.pt'))
            model_graph.eval(); model_linear.eval()
            merged_ret = pd.DataFrame()
            for _, data in enumerate(test_loader):            
                data = data.to(torch.float32).to(self.device)

                zlist=torch.reshape(model_graph(mat_G, data), (1, -1))      
                output_frac = model_linear(zlist)

                partial_ret = (output_frac.cpu().detach().clone().numpy()).reshape((-1, 1))
                partial_ret = pd.DataFrame(partial_ret)
                merged_ret = pd.concat([merged_ret, partial_ret])

            final_ret = pd.concat([final_ret, merged_ret], axis=1)
        final_ret = pd.DataFrame(final_ret.values,index=expression.obs_names,columns=tot_cell_list)
        # for debuging
        if save:
            final_ret.to_csv(f"{out_dir}/{project}_prediction_frac.csv")
        return final_ret
    
    def train(
        self,
        presudo_bulk=None,
        bulk_adata=None,
        cell_list=None,
        sc_adata=None,
        annotation_key = None,
        batch_size = Const.BATCH_SIZE,
        out_dir = "./",
        project_name="",
        data_num=10000,
        batch_effect=True,
        is_st=False
    ):
        """
        out_dir: string, the directory for saving trained models.
        expression: string, needed if `mode` is `training`, the path of the bulk expression file.
        fraction: string, needed if `mode` is `training`, the path of the bulk fraction file.
        marker: string, needed if `mode` is `training`, the path of the gene marker file.
        sc_folder: string, needed if `mode` is `training`, the path of the folder containing single cell reference.
        """
        
        # checking
        if self.mode != Const.MODE_TRAINING:
            raise ValueError("This function can only be used under training mode.")
        
        utils.check_paths(output_folder=out_dir)
        utils.check_paths(output_folder=out_dir+"/plot")

        '''
        if expression.shape[0] != fraction.shape[0]:
            raise ValueError(f"Please check the input, the shape of the expression file {expression.shape} \
                            does not match the one of fraction {fraction.shape}.")
        '''
        tot_cell_list = cell_list
        presudo_bulk_full = presudo_bulk.copy()
        bulk_adata_full = bulk_adata.copy()
        presudo_bulk_full, bulk_adata_full = preprocessing.remove_batch_effect(presudo_bulk_full, bulk_adata_full, out_dir=out_dir, project=project_name,batch_effect=batch_effect)

        if batch_effect:
            loc="batch_effected"
            plots.batch_effect(bulk_adata_full, presudo_bulk_full,out_dir=out_dir+"/plot",title=project_name)
        else:
            loc=None
        
        sample_list = utils.filter_samples(presudo_bulk_full, bulk_adata_full,data_num=data_num,loc=loc)

        presudo_bulk_train = presudo_bulk_full[sample_list,:]

        expression = get.count_data_t(presudo_bulk_train,counts_location=loc)
        #reference = get.count_data_t(bulk_adata_full,counts_location=loc)
        fraction = get.meta(presudo_bulk_train,position_key="obs")
        if torch.backends.mps.is_available():
            for cell in tot_cell_list:
                train_cell_loop_once(cell, expression, fraction,bulk_adata_full, batch_size, sc_adata, out_dir, self.device,annotation_key,project_name,data_num,loc,is_st)
        else:
            for cell in tot_cell_list:
                train_cell_loop_once(cell, expression, fraction, bulk_adata_full, batch_size, sc_adata, out_dir, self.device,annotation_key,project_name,data_num,loc,is_st)
            



import pytest
import os
import sys
import numpy as np
import anndata as ad
import pandas as pd
import cytobulk as ct
import scanpy as sc

@pytest.mark.skip
def test_read_adata(adata_path):
    return sc.read_h5ad(adata_path)

@pytest.mark.parametrize("sc_adata,image_dir,out_dir,project", [(
        "C:/Users/wangxueying/project/CytoBulk/case/he_image/svs/TCGA_LUSC/sub_HTAN_MSK.h5ad",
        "C:/Users/wangxueying/project/CytoBulk/case/he_image/svs/input/TCGA-37-4132_sub2_split",
        "C:/Users/wangxueying/project/CytoBulk/case/he_image/svs/out/TCGA-37-4132_sub2",
        "TCGA-37-4132")])
def test_he_prediction(sc_adata,image_dir,out_dir,project):
    sc_adata = test_read_adata(sc_adata)
    mapping = {
    'plasma cells': 'Plasma Cells',
    'connective tissue': 'Connective Tissue',
    'epithelial': 'Epithelial Cells',
    'neutrophils': 'Neutrophils',
    'lymphocytes': 'Lymphocytes'}

    sc_adata.obs['he_cell_type'] = sc_adata.obs['he_cell_type'].replace(mapping)

    ct.tl.he_mapping(sc_adata = sc_adata,
                     image_dir = image_dir,
                    out_dir =out_dir,
                    project= project,
                    lr_data="C:/Users/wangxueying/project/CytoBulk/case/he_image/svs/input/lrpairs.csv",
                    annotation_key="he_cell_type")




import pandas as pd
from . import model
from .. import preprocessing as pp
from os.path import exists
import json
import time
import scanpy as sc


def _bulk_sc_deconv(bulk_adata, 
                    presudo_bulk, 
                    sc_adata,
                    common_cells,
                    annotation_key = "cell_type", 
                    dataset_name = "",
                    counts_location="batch_effected", 
                    out_dir=".",
                    batch_effect=True,
                    is_st=False):
    """
    Deconvolute the cell type fraction from bulk expression data with single cell dataset as reference.

    Parameters
    ----------
    bulk_adata : anndata.AnnData
        An :class:`~anndata.AnnData` containing the input bulk expression.
    presudo_bulk : anndata.AnnData
        An :class:`~anndata.AnnData` containing the stimulated bulk expression with single cell dataset.
    sc_adata : anndata.AnnData
        An :class:`~anndata.AnnData` containing the single cell expression.
    annotation_key : string
        The `.obs` key where the single cell annotation is stored. : anndata.AnnData.
    marker_data : 
        An :class:`~pandas.dataframe` which columns are cell types, rows are marker gene.
    dataset_name : string.
        The prefix of output file.
    counts_location : string.
        The layer name of the expression in presudo_bulk to train the model.
    out_dir : string, optional
        The path to store the output data.
    **kwargs : parameters in _filter_adata function.
        
    Returns
    -------
    Returns the preprocessed bulk data (adata) , stimualted bulk data and sc data (adata).

    """
    start_t = time.perf_counter()
    deconv = model.GraphDeconv(mode="training")
    #training_data = get.count_data_t(presudo_bulk,counts_location=counts_location)
    #training_fraction = get.meta(presudo_bulk,position_key="obs")
    print("=================================================================================================")
    print('Start to train model.')
    model_dir = out_dir + '/st_model' if is_st else out_dir + '/model'
    deconv.train(out_dir=model_dir,
                presudo_bulk=presudo_bulk,
                bulk_adata=bulk_adata,
                cell_list=common_cells,
                sc_adata = sc_adata,
                annotation_key = "curated_cell_type",
                project_name = dataset_name,
                data_num=bulk_adata.n_obs*5,
                batch_effect=batch_effect,
                is_st=is_st)
    print(f'Time to finish training model: {round(time.perf_counter() - start_t, 2)} seconds')
    print("=================================================================================================")
    '''
    training_data = get.count_data_t(presudo_bulk,counts_location=counts_location)
    training_fraction = get.meta(presudo_bulk,position_key="obs")
    test_data = get.count_data_t(bulk_adata,counts_location=counts_location)
    print("=================================================================================================")
    print('Start to train model.')
    model_dir = out_dir + '/st_model' if is_st else out_dir + '/model'
    deconv.train(out_dir=model_dir,
                expression=training_data,
                fraction=training_fraction,
                marker=marker_dict,
                sc_adata = sc_adata,
                annotation_key = annotation_key,
                is_st=is_st)
    print(f'Time to finish training model: {round(time.perf_counter() - start_t, 2)} seconds')
    print("=================================================================================================")
    '''
    print("=================================================================================================")
    print('Start to predict')
    #test_data = get.count_data_t(bulk_adata)
    start_t = time.perf_counter()
    deconv_result = deconv.fit(
                out_dir=out_dir+'/output',
                expression=bulk_adata,
                cell_list=common_cells,
                sc_adata = sc_adata,
                annotation_key = annotation_key,
                model_folder=model_dir,
                file_dir=model_dir,
                project = dataset_name,
                is_st=is_st)
    print(f'Time to finish deconvolution: {round(time.perf_counter() - start_t, 2)} seconds')
    print("=================================================================================================")
    return deconv_result

def bulk_deconv(bulk_data,
                sc_adata,
                annotation_key,
                marker_list=None,
                rename=None,
                dataset_name="",
                out_dir='.',
                different_source=True,
                cell_list=None,
                scale_factors=100000,
                trans_method="log",
                save = True,
                save_figure=True,
                n_cell=2000,
                **kwargs):
    """
    Deconvolute the cell type fraction from bulk expression data with single cell dataset as reference.
    Reconstruct the bulk data using single cell.

    Parameters
    ----------
    bulk_data : dataframe
        An :class:`~pandas.dataframe` containing the bulk expression data. 
        The first column should be gene symbol, following column should be sample name.
    sc_adata : anndata.AnnData
        An :class:`~anndata.AnnData` containing the single cell expression.
    annotation_key : string
        The `.obs` key where the single cell annotation is stored. : anndata.AnnData.
    marker_data : 
        An :class:`~pandas.dataframe` which columns are cell types, rows are marker gene.
    dataset_name : string.
        The prefix of output file.
    out_dir : string, optional
        The path to store the output data.
    different_source : boolean, optional
        True for single cell and bulk data from the same sample, which means not executing batch effect.
        False for single cell and bulk data from the different samples, which means executing batch effect.
    cell_list : list, optional
        The list indicate the cell type names which need to take into consideration.
    scale_factors : int, optional
        The number of counts to normalize every observation to before computing profiles. If `None`, no normalization is performed. 
    trans_method : string, optional
        What transformation to apply to the expression before computing the profiles. 
        - "log" : log(x+1)
        - `None` : no transformation
    save : boolean, optional
        Whether save the result data during each step. If saving, the processing may be skipped.
    save_figure : boolean, optional
        Whether save figures during preprocessing. eg. scatter plot of pca data.
    mapping_sc : boolean, optional
        Whether reconstruct the bulk data with single cell data.
    n_cell : int, optional
        The number of cells within each bulk.
    **kwargs : 
        Additional keyword arguments forwarded to
        :func:`~cytobulk.preprocessing.qc_bulk_sc`.

        
    Returns
    -------
    Returns the deconvolution result and reconstructed bulk.
    """
    # check the filtered dataset. If exist, skipping preprocessing.
    bulk_ori_adata = bulk_data.copy()
    if exists(f'{out_dir}/filtered/pseudo_bulk_{dataset_name}.h5ad') and exists(f'{out_dir}/filtered/sc_data_{dataset_name}.h5ad') and \
    exists(f'{out_dir}/filtered/bulk_data_{dataset_name}.h5ad') and exists(f'{out_dir}/filtered/cells_{dataset_name}.json'):
        
        pseudo_bulk = sc.read_h5ad(f"{out_dir}/filtered/pseudo_bulk_{dataset_name}.h5ad")
        sc_adata = sc.read_h5ad(f"{out_dir}/filtered/sc_data_{dataset_name}.h5ad")
        bulk_adata = sc.read_h5ad(f"{out_dir}/filtered/bulk_data_{dataset_name}.h5ad")
        with open(f"{out_dir}/filtered/cells_{dataset_name}.json") as json_file:
            common_cell = json.load(json_file)
        annotation_key="curated_cell_type"
    else:
        #preprocessing
        sc_adata, pseudo_bulk, bulk_adata, common_cell,annotation_key = pp.preprocessing(bulk_data,
                                                                        sc_adata,
                                                                        annotation_key,
                                                                        is_st=False,
                                                                        marker_list=marker_list,
                                                                        rename=rename,
                                                                        dataset_name=dataset_name,
                                                                        out_dir=out_dir,
                                                                        different_source=different_source,
                                                                        cell_list=cell_list,
                                                                        scale_factors=scale_factors,
                                                                        trans_method=trans_method,
                                                                        n_sample_each_group=len(bulk_data.obs_names)*5,
                                                                        min_cells_each_group=n_cell,
                                                                        cell_gap_each_group=50,
                                                                        group_number=10,
                                                                        save = save,
                                                                        save_figure=save_figure,
                                                                        **kwargs)
    #deconvolution
    if exists(f'{out_dir}/output/{dataset_name}_prediction_frac.csv'):
        deconv_result = pd.read_csv(f'{out_dir}/output/{dataset_name}_prediction_frac.csv',index_col=0)
    else:
        deconv_result = _bulk_sc_deconv(bulk_adata, 
                                        pseudo_bulk, 
                                        sc_adata, 
                                        common_cell,
                                        annotation_key = annotation_key, 
                                        dataset_name=dataset_name, 
                                        out_dir=out_dir)
    row_sums = deconv_result.sum(axis=1)
    df_normalized = deconv_result.div(row_sums, axis=0)
    bulk_ori_adata.uns['deconv']=df_normalized
    df_normalized.to_csv(f"{out_dir}/output/{dataset_name}_prediction_frac_normalized.csv")
    bulk_ori_adata.write_h5ad(f'{out_dir}/output/{dataset_name}_bulk_adata.h5ad')

    return deconv_result,bulk_ori_adata


def st_deconv(st_adata,
            sc_adata,
            annotation_key,
            marker_list=None,
            rename=None,
            dataset_name="",
            out_dir='.',
            different_source=True,
            cell_list=None,
            scale_factors=10000,
            trans_method="log",
            save = True,
            save_figure=True,
            n_cell=10,
            max_cells=40,
            **kwargs):
    """
    Deconvolute the cell type fraction from spot expression data with single cell dataset as reference.

    Parameters
    ----------
    st_adata : anndata.AnnData
        An :class:`~anndata.AnnData` containing the single cell expression.
        The first column should be gene symbol, following column should be spot name.
    sc_adata : anndata.AnnData
        An :class:`~anndata.AnnData` containing the single cell expression.
    annotation_key : string
        The `.obs` key where the single cell annotation is stored. : anndata.AnnData.
    marker_list : 
        An :class:`~pandas.dataframe` which columns are cell types, rows are marker gene.
    dataset_name : string.
        The prefix of output file.
    out_dir : string, optional
        The path to store the output data.
    different_source : boolean, optional
        True for single cell and bulk data from the same sample, which means not executing batch effect.
        False for single cell and bulk data from the different samples, which means executing batch effect.
    cell_list : list, optional
        The list indicate the cell type names which need to take into consideration.
    scale_factors : int, optional
        The number of counts to normalize every observation to before computing profiles. If `None`, no normalization is performed. 
    trans_method : string, optional
        What transformation to apply to the expression before computing the profiles. 
        - "log" : log(x+1)
        - `None` : no transformation
    save : boolean, optional
        Whether save the result data during each step. If saving, the processing may be skipped.
    save_figure : boolean, optional
        Whether save figures during preprocessing. eg. scatter plot of pca data.
    n_cell : int, optional
        The number of cells within each bulk.
    **kwargs : 
        Additional keyword arguments forwarded to
        :func:`~cytobulk.preprocessing.qc_bulk_sc`.

        
    Returns
    -------
    Returns the deconvolution result and reconstructed st.
    """
    # check the filtered dataset. If exist, skipping preprocessing.
    st_ori_adata = st_adata.copy()
    if exists(f'{out_dir}/filtered/pseudo_bulk_{dataset_name}.h5ad') and exists(f'{out_dir}/filtered/sc_data_{dataset_name}.h5ad') and \
    exists(f'{out_dir}/filtered/bulk_data_{dataset_name}.h5ad') and exists(f'{out_dir}/filtered/cells_{dataset_name}.json'):
        pseudo_st = sc.read_h5ad(f"{out_dir}/filtered/pseudo_bulk_{dataset_name}.h5ad")
        sc_adata = sc.read_h5ad(f"{out_dir}/filtered/sc_data_{dataset_name}.h5ad")
        st_adata = sc.read_h5ad(f"{out_dir}/filtered/bulk_data_{dataset_name}.h5ad")
        with open(f"{out_dir}/filtered/cells_{dataset_name}.json") as json_file:
            common_cell = json.load(json_file)
        annotation_key="curated_cell_type"
    else:
        #preprocessing
        sc_adata, pseudo_st, st_adata,common_cell,annotation_key = pp.preprocessing(st_adata,
                                                                        sc_adata,
                                                                        annotation_key,
                                                                        is_st=True,
                                                                        marker_list=marker_list,
                                                                        rename=rename,
                                                                        dataset_name=dataset_name,
                                                                        out_dir=out_dir,
                                                                        different_source=different_source,
                                                                        cell_list=cell_list,
                                                                        scale_factors=scale_factors,
                                                                        trans_method=trans_method,
                                                                        n_sample_each_group=len(st_adata.obs_names)*6,
                                                                        min_cells_each_group=n_cell,
                                                                        cell_gap_each_group=1,
                                                                        group_number=5,
                                                                        save = save,
                                                                        save_figure=save_figure,
                                                                        **kwargs)
    #deconvolution
    if exists(f'{out_dir}/output/{dataset_name}_prediction_frac.csv'):
        deconv_result = pd.read_csv(f'{out_dir}/output/{dataset_name}_prediction_frac.csv',index_col=0)
    else:
        deconv_result = _bulk_sc_deconv(st_adata, 
                                        pseudo_st, 
                                        sc_adata, 
                                        common_cell,
                                        annotation_key = annotation_key, 
                                        dataset_name=dataset_name, 
                                        out_dir=out_dir,
                                        batch_effect=different_source,
                                        is_st=True)
    threshold = 1 / max_cells
    deconv_result[deconv_result < threshold] = 0
    row_sums = deconv_result.sum(axis=1)
    df_normalized = deconv_result.div(row_sums, axis=0)
    st_ori_adata.uns['deconv']=df_normalized
    df_normalized.to_csv(f"{out_dir}/output/{dataset_name}_prediction_frac_normalized.csv")
    st_ori_adata.write_h5ad(f'{out_dir}/output/{dataset_name}_st_adata.h5ad')
    return df_normalized,st_ori_adata



import numpy as np
import pandas as pd
import math
from scipy.sparse import issparse
from sklearn.preprocessing import StandardScaler
import anndata._core.views
from sklearn.decomposition import PCA
import scanpy as sc
import scipy.spatial as sp

def get_sum(
    X,
    axis,
    dtype=None,
):

    """
    Calculates the sum of a sparse matrix or array-like in a specified axis and
    returns a flattened result.
    
    Parameters
    ----------
    X
        A 2d :class:`~numpy.ndarray` or a `scipy` sparse matrix.
    axis
        The axis along which to calculate the sum
    dtype
        The dtype used in the accumulators and the result.
        
    Returns
    -------
    The flattened sums as 1d :class:`~numpy.ndarray`.
        
    """

    if issparse(X):
        result = X.sum(axis=axis, dtype=dtype).A.flatten()
    else:
        result = X.sum(axis=axis, dtype=dtype)
    

    if isinstance(result, anndata._core.views.ArrayView):
        result = result.toarray()
    
    return result




def pca(X,dimension=2):
    
    """
    Calculates decompositioned data  with 2 dimensions.
    
    Parameters
    ----------
    X
        A :class:`~pd.dataframe` with more than 2 dimension.
    dimension: int
        The number to indicate needed dimension.

        
    Returns
    -------
    The dataframe after dimension reduction with PCA function in sklearn.decomposition.
        
    """

    pca = PCA(n_components=dimension)
    new_data = pca.fit_transform(X.values)

    return pd.DataFrame(
        new_data,
        index=X.index,
        columns=[f'PCA{str(x)}' for x in range(1, dimension + 1)],
    )


def normalization_cpm(adata,scale_factors=None,trans_method=None,layer=None):
    """
    Normalize counts per cell.

    Parameters
    ----------
    scale_factors: int, optional
        After normalization, each observation (cell) has a total count equal to the median 
        of total counts for observations (cells) before normalization.
    trans_method: None or 'log', optional
        If log, Computes X=log(X+1), where log denotes the natural logarithm unless a different base is given.
    layer:
        
    Returns
    -------
    Returns the expression after removing batch effect.

    """
    data = adata.copy()
    if scale_factors is not None:
        sc.pp.normalize_total(data, target_sum=scale_factors)
    if trans_method == 'log':
        sc.pp.log1p(data)
    return data

def normal_center_df(data):
    scaler = StandardScaler()
    scaler.fit(data.values)
    trans_data = scaler.transform(data.values)
    return pd.DataFrame(trans_data,index=data.index,columns=data.columns)

    
def pear(A,B):
    tmp = np.corrcoef(A, B)
    return tmp[0,1] 

def calculate_distance(matrix1,matrix2):
    return (1 - sp.distance.cdist(matrix1, matrix2, 'cosine'))


# Rotation matrix function
def rotate_matrix (x, y, angle, x_shift=0, y_shift=0, units="DEGREES"):
    """
    Rotates a point in the xy-plane counterclockwise through an angle about the origin
    https://en.wikipedia.org/wiki/Rotation_matrix
    :param x: x coordinate
    :param y: y coordinate
    :param x_shift: x-axis shift from origin (0, 0)
    :param y_shift: y-axis shift from origin (0, 0)
    :param angle: The rotation angle in degrees
    :param units: DEGREES (default) or RADIANS
    :return: Tuple of rotated x and y
    """

    # Shift to origin (0,0)
    x = x - x_shift
    y = y - y_shift

    # Convert degrees to radians
    if units == "DEGREES":
        angle = math.radians(angle)

    # Rotation matrix multiplication to get rotated x & y
    yr = (x * math.cos(angle)) - (y * math.sin(angle)) + x_shift
    xr = (x * math.sin(angle)) + (y * math.cos(angle)) + y_shift

    return xr,yr

def convert_pvalue_to_asterisks(pvalue):
    if pvalue <= 0.0001:
        return "****"
    elif pvalue <= 0.001:
        return "***"
    elif pvalue <= 0.01:
        return "**"
    elif pvalue <= 0.05:
        return "*"
    return "ns"


import pytest
import os
import sys
import numpy as np
import anndata as ad
import pandas as pd
import cytobulk as ct
import scanpy as sc
from string import ascii_uppercase

# tested
@pytest.mark.skip
def test_read_adata(adata_path):
    return sc.read_h5ad(adata_path)

@pytest.mark.skip
def test_read_df(data_path):
    return pd.read_csv(data_path,index_col=0,sep='\t')

@pytest.mark.skip
@pytest.mark.parametrize("adata_path", [("C:/Users/wangxueying/project/CytoBulk/case/mouse_mob/output/output/mouse_mob_st_adata.h5ad")])
def test_st_pie_plot(adata_path):
    adata = test_read_adata(adata_path)
    ct.plots.celltype_fraction_pie(adata=adata,
                                   scale_facter_x=5,
                                   scale_factor_y=5,
                                   out_dir="C:/Users/wangxueying/project/CytoBulk/case/mouse_mob/output/output",
                                   rotation_angle=225)

@pytest.mark.skip
@pytest.mark.parametrize("adata_path", [("C:/Users/wangxueying/project/CytoBulk/case/mouse_mob/output/output/mouse_mob_st_adata.h5ad")])
def test_st_cell_heatmap_plot(adata_path):
    adata = test_read_adata(adata_path)
    ct.plots.celltype_fraction_heatmap(adata=adata,
                                       label='OSN',
                                        out_dir="C:/Users/wangxueying/project/CytoBulk/case/mouse_mob/output/output",
                                        rotation_angle=225)
    
@pytest.mark.skip
@pytest.mark.parametrize("adata_path", [("C:/Users/wangxueying/project/CytoBulk/case/mouse_mob_1/output/output/mouse_mob_st_adata.h5ad")])
def test_st_cell_heatmap_plot(adata_path):
    adata = test_read_adata(adata_path)
    ct.plots.paired_violin(adata=adata,
                        label='OSN',
                        gene='Kctd12',
                        out_dir="C:/Users/wangxueying/project/CytoBulk/case/mouse_mob_1/output/output")


@pytest.mark.skip  
@pytest.mark.parametrize("adata_path", [("C:/Users/wangxueying/project/CytoBulk/case/MOB_layer_cut_add_7181/output/reconstructed_mouse_mob_st.h5ad")])
def test_st_reconstruction(adata_path):
    adata = test_read_adata(adata_path)
    ct.plots.reconstruction_corr(adata=adata,
                                out_dir="C:/Users/wangxueying/project/CytoBulk/case/MOB_layer_cut_add_7181/output",
                                rotation_angle=225)

@pytest.mark.skip  
@pytest.mark.parametrize("adata_path", [("C:/Users/wangxueying/project/CytoBulk/case/PDAC/out/output/reconstructed_PDAC_st.h5ad")])
def test_st_reconstruction_pdac(adata_path):
    adata = test_read_adata(adata_path)
    ct.plots.reconstruction_corr(adata=adata,
                                out_dir="C:/Users/wangxueying/project/CytoBulk/case/PDAC/out",
                                rotation_angle=225) 
@pytest.mark.skip    
@pytest.mark.parametrize("adata_path", [("C:/Users/wangxueying/project/CytoBulk/case/10x/sub4/output/reconstructed_sub4_st.h5ad")])
def test_st_reconstruction_10x(adata_path):
    adata = test_read_adata(adata_path)
    ct.plots.reconstruction_corr(adata=adata,
                                out_dir="C:/Users/wangxueying/project/CytoBulk/case/10x/sub4",
                                spot_size=1) 
    
@pytest.mark.skip 
@pytest.mark.parametrize("adata_path", [("C:/Users/wangxueying/project/CytoBulk/case/10x/sub6/output/reconstructed_sub6_st.h5ad")])
def test_st_reconstruction_10x_6(adata_path):
    adata = test_read_adata(adata_path)
    ct.plots.reconstruction_corr(adata=adata,
                                out_dir="C:/Users/wangxueying/project/CytoBulk/case/10x/sub6",
                                spot_size=1) 
    
@pytest.mark.skip 
@pytest.mark.parametrize("adata_path", [("C:/Users/wangxueying/project/CytoBulk/case/10x/sub3/output/reconstructed_sub3_st.h5ad")])
def test_st_reconstruction_10x_3(adata_path):
    adata = test_read_adata(adata_path)
    ct.plots.reconstruction_corr(adata=adata,
                                out_dir="C:/Users/wangxueying/project/CytoBulk/case/10x/sub3",
                                spot_size=1) 
    

@pytest.mark.parametrize("adata_path", [("C:/Users/wangxueying/project/CytoBulk/case/10x/sub6/output/reconstructed_sub6_st.h5ad")])
def test_st_reconstruction_10x_5(adata_path):
    adata = test_read_adata(adata_path)
    ct.plots.reconstruction_corr(adata=adata,
                                out_dir="C:/Users/wangxueying/project/CytoBulk/case/10x/sub6",
                                spot_size=1) 


from sklearn.metrics import mean_squared_error
import pandas as pd
from scipy.stats import pearsonr
from ._read_data import check_paths

def _p_value(data1,data2):
    pccs = pearsonr(data1, data2)
    pval = pccs.pvalue
    stat = '{:.2f}'.format(pccs.statistic)
    return pval,stat


def eval_fraction(df1,df2,out_dir=',',save=True):
    cells=[]
    sig=[]
    mse=[]
    p_value=[]
    cell_name = df1.columns.values.tolist()
    for j in range(len(cell_name)):
        data1 = df1.loc[:,cell_name[j]].values
        data2 = df2.loc[:,cell_name[j]].values
        mse_tmp = mean_squared_error(data1, data2)
        pval,stat = _p_value(data1,data2)
        cells.append(cell_name[j])
        sig.append(float(stat))
        str_pval='none'
        if pval <=0.001:
            str_pval = '***'
        elif 0.001<pval<=0.01:
            str_pval = '**'
        elif 0.01<pval<=0.05:
            str_pval = '*'
        elif str(pval)=="nan":
            str_pval = 'X'
        p_value.append(str_pval)
        mse.append(mse_tmp)
    dict_data = {"cell type":cells,"person correlation":sig,"p_value":p_value,"mse":mse}
    eval_result = pd.DataFrame(dict_data)
    out_dir = check_paths(out_dir+'/output')
    if save:
        eval_result.to_csv(out_dir+"/prediction_eval.txt",sep='\t')
    return eval_result



def eval_comparsion(df1,df2_list,method_name,out_dir=',',save=True):
    cells=[]
    sig=[]
    mse=[]
    p_value=[]
    cell_name = df1.columns.values.tolist()
    methods = []
    for i in range(len(df2_list)):
        for j in range(len(cell_name)):
            data1 = df2_list[i].loc[:,cell_name[j]].values
            data2 = df1.loc[:,cell_name[j]].values
            mse_tmp = mean_squared_error(data1, data2)
            pval,stat = _p_value(data1,data2)
            cells.append(cell_name[j])
            sig.append(float(stat))
            methods.append(method_name[i])
            if pval <=0.001:
                p_value.append('***')
            elif 0.001<pval<=0.01:
                p_value.append('**')
            elif 0.01<pval<=0.05:
                p_value.append('*')
            elif str(pval)=="nan":
                p_value.append('X')
            else:
                p_value.append('X')
            mse.append(mse_tmp)
    dict_data = {"cell type":cells,"method":methods,"person correlation":sig,"p_value":p_value,"mse":mse}
    eval_result = pd.DataFrame(dict_data)
    out_dir = check_paths(out_dir+'/output')
    if save:
        eval_result.to_csv(out_dir+"/fraction_eval.txt",sep='\t')
    return eval_result


def eval_each_sample_mse(df1,df2,method_name,out_dir=',',save=True):
    cells=[]
    mse=[]
    samples=[]
    cell_name = df1.columns.values.tolist()
    methods = []
    sample_name = df1.index.tolist()
    df2 = df2.loc[sample_name,:]
    for i in range(len(df1)):
        for j in range(len(cell_name)):
            data1 = df2.loc[i,cell_name[j]].values
            data2 = df1.loc[i,cell_name[j]].values
            mse_tmp = mean_squared_error(data1, data2)
            samples.append(sample_name[i])
            cells.append(cell_name[j])
            methods.append(method_name)
            mse.append(mse_tmp)
    dict_data = {"sample_name":samples,"cell_type":cells,"method":methods,"mse":mse}
    eval_result = pd.DataFrame(dict_data)
    if save:
        out_dir = check_paths(out_dir+'/output')
        eval_result.to_csv(out_dir+"/fraction_eval.txt",sep='\t')
    return eval_result